/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __TUPLE_H_
#define __TUPLE_H_

/*
 * It represent an ungapped alignment it term of 
 * "diagonal" and "occurrence" on the right sequence
 * 
 * - "left" is the "lenght" of already matching hits on the
 * same diagonal (a tuple is always on one single diagonal).
 * - "lastSeed" is the seed number (that enables to compute
 *   the right span ...)
 * - "bitweight" and "overlappedhits" are only used in the assemble/align
 *    algorithms
 */

typedef struct _tuple {
    struct _tuple *next;
    int occurrence;
    int diagonal;
    int leftsize;
    unsigned int   bitweight;
    unsigned short overlappedhits;
    unsigned short lastSeed;
} tuple;

typedef struct _tuplelist {
    struct _tuplelist *next;
    struct _tuple *first_tuple;
} tuplelist;


/*
 * MA : "Memorized Alignment" 
 *       represents an alignment in term 
 *       of left/rigth (query/text) sequences and
 *       begining-ending position on each sequence
 *       
 *       In order to quicky align sequences, 
 *       chain of tuples is kept as a "backbone"
 *       for the alignment
 */

typedef struct _MA {

  struct _MA * next;
  struct _tuple *first_tuple;
  
  /* position of memorized alignment */
  int left_pos_begin;
  int right_pos_begin;
  int left_pos_end;
  int right_pos_end;
  
  /* blastscore */
  int blastscore;
  float entropy;
  float mutual;
  
  /* chunk used on the second sequence */
  int i_chunk;
  /* reverse (boolean) */
  unsigned char reverse;

} MA;



/*  
 *      LEFT SEQUENCE : QUERY                               RIGHT SEQUENCE : TEXT
 *
 *
 *
 *
 *  TBL_POS(t1)  TEL_POS(t1)                            TBR_POS(t1)  TER_POS(t1) 
 *     |           |                                        |            |
 *     |           |                                        |            |
 *-----[=========#=]-----[====#=]---------------------------[==========#=][====#=]-------------
 *               |            |                                        |       | 
 *               |            |                                        |       | 
 *          TL_POS(t1)  TL_POS(t2)                              TR_POS(t1)  TR_POS(t2)
 *
 *	
 *
 *
 *
 *
 *        TSIZE(t1) 
 *     <----------->
 *     .           .
 *     .           .
 *-----[=========#=]-----[====#=]---------------------------[==========#=][====#=]-------------
 *     .           .     .      .                           .            ..      . 
 *     .           .     .      .                           .            ..      .                    
 *     .           .     .      .                           .            ..      .  
 *     .           <----->      .                           .            ..      .              
 *     .         TGAP_L(t1,t2)  .                           .           -><-     .                         
 *     .                        .                           .     TGAP_R(t1,t2)  .             
 *     .                        .                           .                    .       
 *     <------------------------>                           .                    .                                              
 *            TSEG_L(t1,t2)                                 <-------------------->
 *                                                               TSEG_R(t1,t2) 
 */


#include "global_var.h"
#include "threads.h"

/* position on the text (R as right) */
#define TR_POS(t) ((t)->occurrence)

/* position on the query (L as left) 
 *   TL_POS should be modified according to "left_correction" 
 *   whith can be either 0 or keysize
 *
 */
#define TL_POS1(t) ((t)->occurrence - (t)->diagonal)
#define TL_POS(t)  ((t)->occurrence - (t)->diagonal + left_correction)


/* position begin "B" and end "E" on each Left and Rigth seed */

#define TBR_POS(t)  (TR_POS(t) - (t)->leftsize)
#define TBL_POS(t)  (TL_POS(t) - (t)->leftsize)

/* End position + 1 !!! */
#define TER_POS(t)  (TR_POS(t) + gp_seeds_span[t->lastSeed] )
#define TEL_POS(t)  (TL_POS(t) + gp_seeds_span[t->lastSeed] )


/* length */

/* length of one segment (t1<t2) between 2 seeds (including seeds size) */
#define TSEG_L(t1,t2) (TL_POS1(t2) - TL_POS1(t1) + (t1)->leftsize + gp_seeds_span[t2->lastSeed])
#define TSEG_R(t1,t2) (TR_POS(t2)  - TR_POS(t1)  + (t1)->leftsize + gp_seeds_span[t2->lastSeed])

/* length of one gap     (t1<t2) between 2 seeds (excluding seeds size) */
#define TGAP_L(t1,t2) (TL_POS1(t2) - TL_POS1(t1) - (t2)->leftsize - gp_seeds_span[t1->lastSeed])
#define TGAP_R(t1,t2) (TR_POS(t2)  - TR_POS(t1)  - (t2)->leftsize - gp_seeds_span[t1->lastSeed])

/* length of one "unsized" gap  (t1<t2) between 2 seeds (seeds size) */
#define TUSGAP_L(t1,t2) (TL_POS1(t2) - TL_POS1(t1) - gp_seeds_span[t1->lastSeed])
#define TUSGAP_R(t1,t2) (TR_POS(t2)  - TR_POS(t1)  - gp_seeds_span[t1->lastSeed])

/* full seed size */
#define TSIZE(t)  ((t)->leftsize + gp_seeds_span[t->lastSeed])


/*
 * [1] tuple/tuplelists/ma (memorized alignments) management
 */


/* Build a tuple */
tuple     * CreateTuple(int occurrence, int diagonal, int seed);
/* Build an empty tuple list */
tuplelist * CreateTupleList(); 
/* Build a Memorized alignment */
MA        * CreateMA(int left_pos_begin, int right_pos_begin,
		     int left_pos_end, int right_pos_end,
		     tuplelist * tuple_list,
		     tuple * first_tuple_aligned,
		     tuple * last_tuple_aligned, 
		     int blastscore, double entropy,
		     Feature * feature
		     );
/* Free a single tuple */
void FreeTuple(tuple * t);
/* Free a list of tuples (based on next chain) */
void FreeTuples(tuple * t);
/* Free a complete MA and its tuple list */
void FreeMA(MA * ma);

/* Free a tuplelist pointed out by p_tl (and its list of tuple),
 * you have to pass "previous" tuplelist element and also a "last_tl"
 * ---
 * Specific function only used in "align" 
 */
void FreeTupleList(tuplelist ** p_tl, tuplelist ** p_tl_prev,tuplelist **last_tl);	
/* Cut and paste the interestion chain of "tl" ("first,last") into "ma" */
void KeepTupleList(tuplelist * tl, tuple * tfirst, tuple * tlast, MA * ma);	
/* Left correction is applied */
void LeftCorrection_MA(MA * ma, int left_correction);



/*
 * [2] Sorting/filtering functions
 */

/* sorting functions : for one MA */
typedef int (SortCrit) (MA *);
int SortCriterionScore(MA * ma)            ;
int SortCriterionEntropy(MA * ma)          ;
int SortCriterionMutual(MA * ma)           ;
int SortCriterionScoreWithEntropy(MA * ma) ;
int SortCriterionQueryBegin(MA * ma)       ;

/* recursive function */
void ListSort_MAList(MA ** p_firstMA, MA ** p_lastMA, int mask /* 0x40000000 */ );
/* main sorting function called by one thread */	
void Sort_MAList(Feature * f);
/* merge results provided by several threads */
MA * MergeSort_MALists(Feature * f1, Feature * f2);


/* filtering functions : for one MA */
int EntropyAndScoreFilter_MA(Feature * feature, MA * ma, char *dataquery,  char *datachunktext);
/* main filtering function called by one thread */
int EntropyAndScoreFilter_MAList(Feature * feature, char *dataquery, char *datatext, int *datatext_start);



/*
 * [3] Display (provided for debugging)
 */
int DisplayListTupleList(tuplelist * tuplelist);
int DisplayTupleList(tuplelist * tuplelist);
int DisplayListTuple(tuplelist * tuple);
int DisplayTuple(tuple * tuple);
int DisplayMA(MA * ma);
int DisplayListMA (MA * ma);

#endif
